"""Install files needed for building sites."""

import argparse
import importlib.resources
from pathlib import Path
import sys


def main(opt):
    """Main driver."""
    src = importlib.resources.files(__name__.split(".")[0]) / "data"
    available = {
        filename: opt.dst / filename.relative_to(src)
        for filename in src.glob("**/*")
        if filename.is_file()
    }
    if opt.only:
        available = {
            filename: path
            for filename, path in available.items()
            if filename.relative_to(src) in opt.only
        }

    exists = {str(dst) for dst in available.values() if dst.exists()}
    if exists and (not opt.force):
        print(
            f"not overwriting {', '.join(sorted(exists))} (use --force)",
            file=sys.stderr,
        )
        sys.exit(1)

    opt.dst.mkdir(parents=True, exist_ok=True)
    for src, dst in available.items():
        dst.parent.mkdir(parents=True, exist_ok=True)
        dst.write_bytes(src.read_bytes())


def construct_parser(parser):
    """Parse command-line arguments."""
    parser.add_argument("--dst", type=Path, default=".", help="destination directory")
    parser.add_argument("--force", action="store_true", help="force overwrite")
    parser.add_argument(
        "--only", type=Path, nargs="+", help="only install specific files"
    )



"""Build HTML site from source files."""

import argparse
from pathlib import Path
import re
import sys

from bs4 import BeautifulSoup
from jinja2 import Environment, FileSystemLoader
from markdown import markdown
import tomli


BOILERPLATE = {
    "README.md": Path(""),
    "CODE_OF_CONDUCT.md": Path("conduct"),
    "CONTRIBUTING.md": Path("contrib"),
    "LICENSE.md": Path("license"),
}
GLOSSARY_MD = re.compile(r'^<span\s+id="(.+?)"\s*>(.+?)</span>', re.MULTILINE)
MARKDOWN_EXTENSIONS = ["attr_list", "def_list", "fenced_code", "md_in_html", "tables"]


def main(opt):
    """Main driver."""
    opt.settings = _load_config(opt.config)
    opt._links = opt.links.read_text() if opt.links else None
    files = _find_files(opt)
    markdown, others = _separate_files(files)
    opt._glossary = _load_glossary(markdown)
    opt.dst.mkdir(parents=True, exist_ok=True)

    env = Environment(loader=FileSystemLoader(opt.templates))
    context = _get_context_from_readme(opt)
    _handle_markdown(opt, env, context, markdown)
    _handle_others(opt, others)


def construct_parser(parser):
    """Parse command-line arguments."""
    parser.add_argument(
        "--config", default=Path("pyproject.toml"), help="configuration file"
    )
    parser.add_argument(
        "--dst", type=Path, default=Path("docs"), help="output directory"
    )
    parser.add_argument("--links", type=Path, default=None, help="links file")
    parser.add_argument("--src", type=Path, default=Path("."), help="source directory")
    parser.add_argument(
        "--templates", type=Path, default="templates", help="templates directory"
    )


def _create_figure_numbers(dest, doc):
    """Modify figures, building key-to-number lookup."""
    seen = {}
    for i, figure in enumerate(doc.select("figure")):
        fig_num = i + 1
        if "id" not in figure.attrs:
            _warn(f"figure {fig_num} in {dest} has no ID")
            continue
        if not figure["id"].startswith("f:"):
            _warn(f"figure {fig_num} in {dest} does not start with 'f:'")
            continue
        captions = figure.select("figcaption")
        if len(captions) != 1:
            _warn(f"figure {fig_num} in {dest} has missing/too many captions")
            continue
        caption = captions[0]
        seen[figure["id"]] = fig_num
        caption.insert(0, f"Figure {fig_num}: ")
    return seen


def _create_table_numbers(dest, doc):
    """Modify tables, building key-to-number lookup."""
    seen = {}
    for i, div in enumerate(doc.select("div[data-table-id]")):
        tbl_num = i + 1
        if not div["data-table-id"].startswith("t:"):
            _warn(f"table {tbl_num} in {dest} does not start with 't:'")
            continue

        if "data-table-caption" not in div.attrs:
            _warn(f"table {tbl_num} in {dest} does not have data-table-caption")
            continue
        caption = div["data-table-caption"]

        tables = div.select("table")
        if len(tables) != 1:
            _warn(f"table {tbl_num} in {dest} does not contain table")
            continue

        table = tables[0]
        table["id"] = div["data-table-id"]
        caption_node = doc.new_tag("caption")
        caption_node.string = caption
        table.append(caption_node)

        seen[div["data-table-id"]] = tbl_num

    return seen


def _do_bibliography_links(opt, dest, doc):
    """Handle 'b:' bibliography links."""
    for node in doc.select("a[href]"):
        if not node["href"].startswith("b:"):
            continue
        assert node["href"].count(":") == 1
        key = node["href"].split(":")[1]
        node.string = key
        node["href"] = _make_root_prefix(opt, dest) + f"bibliography/#{key}"


def _do_figures(opt, dest, doc):
    """Insert figure numbers."""
    seen = _create_figure_numbers(dest, doc)
    _update_cross_references(dest, doc, seen, "figure", "#f:", "Figure")


def _do_glossary_links(opt, dest, doc):
    """Handle 'g:' glossary links."""
    for node in doc.select("a[href]"):
        if not node["href"].startswith("g:"):
            continue
        assert node["href"].count(":") == 1
        key = node["href"].split(":")[1]
        node["href"] = _make_root_prefix(opt, dest) + f"glossary/#{key}"


def _do_glossary_terms(opt, dest, doc):
    """Fill in <p id="terms"></p> if present and terms defined."""
    targets = doc.select("p#terms")
    if len(targets) == 0:
        return
    if len(targets) > 1:
        _warn(f"terms paragraph appears multiple times in {dest}")
        return
    target = targets[0]

    found = {
        node["href"].split("#")[-1]: node["href"]
        for node in doc.select("a[href]")
        if "/glossary/#" in node["href"]
    }
    if not found:
        target.decompose()
        return

    entries = [(key, opt._glossary.get(key, "UNDEFINED")) for key in found.keys()]
    entries.sort(key=lambda item: item[1])
    target.append("Terms defined: ")
    for i, (key, term) in enumerate(entries):
        tag = doc.new_tag("a", href=found[key])
        tag.string = term
        if i > 0:
            target.append(", ")
        target.append(tag)


def _do_markdown_links(opt, dest, doc):
    """Handle ./SOMETHING.md links."""
    for node in doc.select("a[href]"):
        if not node["href"].endswith(".md"):
            continue
        target = str(Path(node["href"]).name)
        if target not in BOILERPLATE:
            _warn(f"unknown Markdown link {node['href']} for {dest}")
            continue
        node["href"] = f"@/" if target == "README.md" else f"@/{BOILERPLATE[target]}/"


def _do_pre_code_classes(opt, dest, doc):
    """Add language classes to <pre> elements."""
    for node in doc.select("pre>code"):
        cls = node.get("class", [])
        node.parent["class"] = node.parent.get("class", []) + cls


def _do_root_links(opt, dest, doc):
    """Handle '@/' links."""
    prefix = _make_root_prefix(opt, dest)
    targets = (
        ("a[href]", "href"),
        ("img[src]", "src"),
        ("link[href]", "href"),
        ("script[src]", "src"),
    )
    for selector, attr in targets:
        for node in doc.select(selector):
            if node[attr].startswith("@/"):
                node[attr] = node[attr].replace("@/", prefix)


def _do_tables(opt, dest, doc):
    """Insert figure numbers."""
    seen = _create_table_numbers(dest, doc)
    _update_cross_references(dest, doc, seen, "table", "#t:", "Table")


def _do_title(opt, dest, doc):
    """Make sure title element is filled in."""
    if doc.title is None:
        _warn(f"{dest} does not have <title> element")
        return
    try:
        doc.title.string = doc.h1.get_text()
    except Exception:
        _warn(f"{dest} lacks H1 heading")


def _find_files(opt):
    """Collect all interesting files."""
    return [path for path in opt.src.glob("**/*.*") if _is_interesting_file(opt, path)]


def _fix_url(url):
    assert url.startswith("./"), f"bad README internal url {url}"
    return url.replace("./", "@/", 1)


def _get_context_from_readme(opt):
    """Get values for template expansion from README.md"""
    md = (opt.src / "README.md").read_text()
    html = markdown(md, extensions=MARKDOWN_EXTENSIONS)
    doc = BeautifulSoup(html, "html.parser")
    title = doc.select_one("h1").decode_contents()
    lessons = {
        _fix_url(node["href"]): node.decode_contents()
        for node in doc.select("div#lessons")[0].select("a[href]")
    }
    appendices = {
        _fix_url(node["href"]): node.decode_contents()
        for node in doc.select("div#appendices")[0].select("a[href]")
    }

    page_order = [
        _normalize_for_nav(url)
        for url in list(lessons.keys()) + list(appendices.keys())
    ]

    normalized_lessons = {_normalize_for_nav(k): v for k, v in lessons.items()}
    normalized_appendices = {_normalize_for_nav(k): v for k, v in appendices.items()}
    all_pages = {**normalized_lessons, **normalized_appendices}

    # Create navigation lookup table
    nav_lookup = {}
    for i, page_url in enumerate(page_order):
        nav_info = {}
        if i > 0:
            prev_url = page_order[i - 1]
            nav_info["prev"] = {"url": prev_url, "title": all_pages[prev_url]}
        if i < len(page_order) - 1:
            next_url = page_order[i + 1]
            nav_info["next"] = {"url": next_url, "title": all_pages[next_url]}
        nav_lookup[page_url] = nav_info

    return {
        "title": title,
        "lessons": lessons,
        "appendices": appendices,
        "page_order": page_order,
        "all_pages": all_pages,
        "nav_lookup": nav_lookup,
    }


def _get_prev_next_context(opt, context, dest):
    """Calculate previous and next page context for navigation."""
    # Convert destination path to URL format for comparison
    relative_dest = dest.relative_to(opt.dst)
    if relative_dest.name == "index.html":
        if str(relative_dest.parent) == ".":
            current_url = "@/"
        else:
            current_url = "@/" + str(relative_dest.parent) + "/"
    else:
        current_url = "@/" + str(relative_dest.with_suffix("")) + "/"

    nav_lookup = context.get("nav_lookup", {})
    nav_info = nav_lookup.get(current_url, {})

    return {
        "prev_page": nav_info.get("prev"),
        "next_page": nav_info.get("next"),
    }


def _handle_markdown(opt, env, context, files):
    """Handle Markdown files."""
    for source in files:
        dest = _make_output_path(opt, source)
        html = _render_markdown(opt, env, context, source, dest)
        dest.write_text(html)


def _handle_others(opt, files):
    """Handle copy-only files."""
    for source in files:
        dest = _make_output_path(opt, source)
        content = source.read_bytes()
        dest.write_bytes(content)


def _is_interesting_file(opt, path):
    """Is this file worth copying over?"""
    relative = path.relative_to(opt.src)
    if not path.is_file():
        return False
    if str(relative).startswith("."):
        return False
    if path.samefile(opt.config):
        return False
    if path.is_relative_to(opt.dst):
        return False
    if path.is_relative_to(opt.templates):
        return False

    skips = opt.settings["skips"]
    for s in skips:
        if s.endswith("/**") and relative.is_relative_to(s.replace("/**", "")):
            return False
        if path.match(s):
            return False

    return True


def _load_config(filepath):
    """Read configuration data."""
    config = tomli.loads(filepath.read_text())

    if ("tool" not in config) or ("mccole" not in config["tool"]):
        _warn(f"configuration file {filepath} does not have 'tool.mccole'")
        return {"skips": set()}

    config = config["tool"]["mccole"]
    config["skips"] = set(config["skips"]) if "skips" in config else set()

    overlap = set(BOILERPLATE.keys()) & config["skips"]
    if overlap:
        _warn(
            f"overlap between skips and renames in config {filepath}: {', '.join(sorted(overlap))}"
        )

    return config


def _load_glossary(markdown_filenames):
    """Get key:term pairs from glossary."""
    paths = [p for p in markdown_filenames if "glossary/index.md" in str(p)]
    if len(paths) == 0:
        _warn(f"no glossary found")
        return {}
    elif len(paths) > 1:
        _warn(f"multiple glossary files")
        return {}
    content = paths[0].read_text()
    return {m[0]: m[1] for m in GLOSSARY_MD.findall(content)}


def _make_output_path(opt, source):
    """Build output path."""
    source_str = str(source.name)
    temp = BOILERPLATE[source_str] / "index.md" if source_str in BOILERPLATE else source
    temp = temp.with_suffix("").with_suffix(".html") if temp.suffix == ".md" else temp
    temp = opt.src / temp
    result = opt.dst / temp.relative_to(opt.src)
    result.parent.mkdir(parents=True, exist_ok=True)
    return result


def _make_root_prefix(opt, path):
    """Create prefix to root for path."""
    relative = path.relative_to(opt.dst)
    depth = len(relative.parents) - 1
    assert depth >= 0
    return "./" if (depth == 0) else "../" * depth


def _normalize_for_nav(url):
    """Convert README URLs to final page URLs for navigation"""
    if not url.endswith(".md"):
        return url
    filename = url.split("/")[-1]
    if filename in BOILERPLATE:
        if filename == "README.md":
            return "@/"
        else:
            return f"@/{BOILERPLATE[filename]}/"


def _render_markdown(opt, env, context, source, dest):
    """Convert Markdown to HTML."""
    content = source.read_text()
    if opt._links:
        content += "\n" + opt._links
    template_name = "slides.html" if source.name == "slides.md" else "page.html"
    template = env.get_template(template_name)
    raw_html = markdown(content, extensions=MARKDOWN_EXTENSIONS)

    # Calculate prev/next navigation
    page_context = dict(context)
    page_context.update(_get_prev_next_context(opt, context, dest))

    rendered_html = template.render(content=raw_html, **page_context)

    doc = BeautifulSoup(rendered_html, "html.parser")
    for func in [
        _do_bibliography_links,
        _do_figures,
        _do_glossary_links,
        _do_glossary_terms,
        _do_markdown_links,
        _do_pre_code_classes,
        _do_root_links,
        _do_tables,
        _do_title,
    ]:
        func(opt, dest, doc)

    return str(doc)


def _separate_files(files):
    """Divide files into categories."""
    markdown = [path for path in files if path.suffix == ".md"]
    others = [path for path in files if path.suffix != ".md"]
    return markdown, others


def _update_cross_references(dest, doc, seen, kind, prefix, caption):
    """Modify cross-references to figures."""
    for ref in doc.select("a[href]"):
        if not ref["href"].startswith(prefix):
            continue
        key = ref["href"][1:]
        if key not in seen:
            _warn(f"cannot resolve {kind} reference {ref['href']} in {dest}")
            continue
        ref.string = f"{caption} {seen[key]}"


def _warn(msg):
    """Print warning."""
    print(msg, file=sys.stderr)



"""Check generated HTML."""

import argparse
from collections import defaultdict
from pathlib import Path
import re
import sys

from bs4 import BeautifulSoup
from html5validator.validator import Validator

RE_FIGURE_CAPTION = re.compile(r"^Figure\s+\d+:")
RE_MD_CODE_PATTERNS = [
    re.compile(r"^```.+?^```", re.DOTALL + re.MULTILINE),
    re.compile(r"`.+?`", re.DOTALL + re.MULTILINE),
]
RE_MD_LINK_DEF = re.compile(r"^\[(.+?)\]:\s+.+?\s*$", re.MULTILINE)
RE_MD_LINK_REF = re.compile(r"\[.+?\]\[(.+?)\]", re.DOTALL | re.MULTILINE)
SECTIONS = [
    ("lessons", "span#nav-lessons", "div#lessons"),
    ("extras", "span#nav-appendices", "div#appendices"),
]


def main(opt):
    """Main driver."""
    opt._links = opt.links.read_text() if opt.links else None
    _check_markdown(opt)
    _check_html(opt)


def construct_parser(parser):
    """Parse command-line arguments."""
    parser.add_argument("--dst", type=Path, default="docs", help="output directory")
    parser.add_argument("--html", action="store_true", help="validate HTML")
    parser.add_argument("--links", type=Path, default=None, help="links file")
    parser.add_argument("--src", type=Path, default=".", help="source directory")


def _check_html(opt):
    """Check generated HTML."""
    html_paths = opt.dst.glob("**/*.html")
    html_pages = {
        path: BeautifulSoup(path.read_text(), "html.parser") for path in html_paths
    }
    for func in [
        _do_compare_readme_section_titles,
        _do_compare_template_readme,
        _do_exercise_titles,
        _do_figure_structure,
        _do_glossary_redefinitions,
        _do_single_h1,
        _do_table_structure,
        lambda o, p: _do_special_links(o, p, "bibliography"),
        lambda o, p: _do_special_links(o, p, "glossary"),
    ]:
        func(opt, html_pages)
    if opt.html:
        _do_html_validation(opt, html_pages)


def _check_markdown(opt):
    """Check source Markdown pages."""
    top_level = {path: path.read_text() for path in opt.src.glob("*.md")}
    sub_level = {path: path.read_text() for path in opt.src.glob("*/*.md")}
    markdown_pages = top_level | sub_level
    for func in [
        _do_unresolved_markdown_links,
    ]:
        func(opt, markdown_pages)


def _do_compare_readme_section_titles(opt, html_pages):
    """Make sure section titles in main page match those in sub-pages."""
    readme_path = opt.dst / "index.html"
    readme = BeautifulSoup(readme_path.read_text(), "html.parser")
    for kind, nav_selector, _ in SECTIONS:
        nav = readme.select(nav_selector)
        if not _require(
            len(nav) == 1, f"{readme_path} missing or multiple {nav_selector}"
        ):
            continue
        nav = nav[0]
        for node in nav.select("a"):
            expected = node.get_text()
            sub_path = opt.dst / node["href"] / "index.html"
            if not _require(
                sub_path in html_pages, f"README sub-path {sub_path} not found"
            ):
                continue
            actual = html_pages[sub_path].select("h1")
            if not _require(len(actual) == 1, f"missing/multiple H1 in {sub_path}"):
                continue
            actual = actual[0].get_text()
            _require(
                expected == actual,
                f"title mis-match between README {repr(expected)} and {sub_path} {repr(actual)}",
            )


def _do_compare_template_readme(opt, html_pages):
    """Compare tables of contents in template and README.md"""
    path = opt.dst / "index.html"
    readme = html_pages[path]
    for kind, nav_selector, body_selector in SECTIONS:
        nav = readme.select(nav_selector)
        if not _require(len(nav) == 1, f"{path} missing or multiple {nav_selector}"):
            continue
        nav = nav[0]
        body = readme.select(body_selector)
        if not _require(len(body) == 1, f"{path} missing or multiple {body_selector}"):
            continue
        body = body[0]

        nav_entries = {node["href"]: node.string for node in nav.select("a[href]")}
        body_entries = {node["href"]: node.string for node in body.select("a[href]")}
        difference = set(nav_entries.keys()) ^ set(body_entries.keys())
        _require(
            not difference,
            f"mis-match in README and nav paths: {', '.join(sorted(difference))}",
        )

        common = set(nav_entries.keys()) & set(body_entries.keys())
        difference = [key for key in common if nav_entries[key] != body_entries[key]]
        _require(
            not difference,
            f"mis-match in README and nav paths text: {' '.join(sorted(difference))}",
        )


def _do_exercise_titles(opt, html_pages):
    """Check that exercise sections have one <h3> heading."""
    for path, doc in html_pages.items():
        for section in doc.select("section.exercise"):
            headings = section.select("h1, h2, h3, h4, h5, h6")
            _require(
                len(headings) == 1,
                f"exercise in {path} has missing/multiple heading(s)",
            )
            if len(headings) == 1:
                _require(
                    headings[0].name.lower() == "h3",
                    f"exercise in {path} has {headings[0].name} instead of h3",
                )


def _do_figure_structure(opt, html_pages):
    """Check that all figures have IDs and captions."""
    for path, doc in html_pages.items():
        for figure in doc.select("figure"):
            _require("id" in figure.attrs, f"figure missing 'id' in {path}")
            captions = figure.select("figcaption")
            if not _require(
                len(captions) == 1, f"figure missing/extra caption in {path}"
            ):
                continue
            text = captions[0].string
            _require(
                RE_FIGURE_CAPTION.match(text),
                f"badly-formatted figure caption '{text}' in {path}",
            )


def _do_glossary_redefinitions(opt, html_pages):
    """Check for glossary terms that are defined more than once."""
    seen = defaultdict(list)
    for path, doc in html_pages.items():
        for node in doc.select("a[href]"):
            if "/glossary/#" in node["href"]:
                key = node["href"].split("#")[-1]
                seen[key].append(path)
    for key, values in seen.items():
        _require(
            len(values) == 1,
            f"glossary entry '{key}' defined in {', '.join(sorted(str(v) for v in values))}",
        )


def _do_html_validation(opt, html_pages):
    """Validate generated HTML."""
    validator = Validator()
    errors = validator.validate(list(html_pages.keys()))


def _do_single_h1(opt, html_pages):
    """Check that each page has exactly one <h1>."""
    for path, doc in html_pages.items():
        titles = doc.find_all("h1")
        _require(len(titles) == 1, f"{len(titles)} H1 elements found in {path}")


def _do_special_links(opt, html_pages, stem):
    """Check specially-formatted links."""
    source = opt.dst / stem / "index.html"
    if not _require(source in html_pages, f"{source} not found"):
        return

    main = html_pages[source].select("main")
    if not _require(len(main) == 1, f"missing or multiple <main> in {source}"):
        return

    main = main[0]
    defined = {node["id"] for node in main.select("span") if node.has_attr("id")}

    base = f"{stem}/#"
    used = set()
    for path, doc in html_pages.items():
        here = {
            node["href"].split("#")[-1]
            for node in doc.select("a[href]")
            if base in node["href"]
        }
        used |= here
        unknown = here - defined
        _require(
            len(unknown) == 0,
            f"unknown {stem} keys in {path}: {', '.join(sorted(unknown))}",
        )

    unused = defined - used
    _require(len(unused) == 0, f"unused {stem} keys: {', '.join(sorted(unused))}")


def _do_table_structure(opt, html_pages):
    """Check that all tables have proper structure and IDs."""
    for path, doc in html_pages.items():
        table_divs = doc.select("div[data-table-id], div[data-table-caption]")
        for div in table_divs:
            if not _require(
                "data-table-id" in div.attrs,
                f"table div missing 'data-table-id' in {path}",
            ):
                continue
            table_id = div["data-table-id"]
            _require(
                table_id.startswith("t:"),
                f"table ID '{table_id}' does not start with 't:' in {path}",
            )
            _require(
                "data-table-caption" in div.attrs,
                f"table div missing 'data-table-caption' in {path}",
            )
            tables = div.select("table")
            _require(
                len(tables) == 1,
                f"table div should contain exactly one table, found {len(tables)} in {path}",
            )


def _do_unresolved_markdown_links(opt, markdown_pages):
    """Check for [text][key] that doesn't resolve and for unused link definitions."""
    general = (
        set()
        if opt._links is None
        else set(m for m in RE_MD_LINK_DEF.findall(opt._links))
    )
    seen = set()
    for path, md in markdown_pages.items():
        for pat in RE_MD_CODE_PATTERNS:
            md = pat.sub("", md)
        referenced = set(m for m in RE_MD_LINK_REF.findall(md))
        in_page = set(m for m in RE_MD_LINK_DEF.findall(md))

        unknown = (referenced - in_page) - general
        _require(
            not unknown,
            f"undefined link reference(s) in {path}: {', '.join(sorted(unknown))}",
        )

        unused = in_page - referenced
        _require(
            not unused,
            f"unused link definition(s) in {path}: {', '.join(sorted(unused))}",
        )

        seen |= referenced & general

    unused = general - seen
    _require(
        not unused, f"unused global link definition(s): {', '.join(sorted(unused))}"
    )


def _require(cond, msg):
    """Check and report."""
    if not cond:
        print(msg, file=sys.stderr)
    return cond



"""Test that sites build."""

import argparse

import pytest

from mccole.build import main as build, construct_parser, _load_glossary
from .conftest import README, make_fs, read_doc


TABLE_MD = """\
# Title

<div markdown="1" data-table-id="t:tbl" data-table-caption="Table Title">
| left | right |
| ---- | ----- |
| 1    | 2     |
| 3    | 4     |
</div>
"""


def test_construct_parser_with_default_values():
    parser = argparse.ArgumentParser()
    construct_parser(parser)
    opt = parser.parse_args([])
    assert all(hasattr(opt, key) for key in ["config", "dst", "src", "templates"])


def test_with_minimal_files(bare_fs, build_opt, minimal_dst_top_level):
    build(build_opt)
    dst = bare_fs / build_opt.dst
    assert dst.is_dir()
    assert set(dst.iterdir()) == minimal_dst_top_level


def test_with_single_plain_markdown_file_creates_output_file(bare_fs, build_opt):
    make_fs({bare_fs / build_opt.src / "test.md": "# Title\nbody"})
    build(build_opt)

    expected = bare_fs / build_opt.dst / "test.html"
    assert expected.is_file()

    doc = read_doc(expected)
    assert doc.title.string == "Title"
    paragraphs = doc.find_all("p")
    assert len(paragraphs) == 1 and paragraphs[0].string == "body"


def test_does_not_copy_dot_files(bare_fs, build_opt, minimal_dst_top_level):
    make_fs(
        {
            bare_fs / build_opt.src / ".gitignore": "content",
        }
    )
    build(build_opt)
    assert set((bare_fs / build_opt.dst).iterdir()) == minimal_dst_top_level


def test_does_not_copy_dot_dirs(bare_fs, build_opt, minimal_dst_top_level):
    (bare_fs / build_opt.src / ".settings").mkdir()
    build(build_opt)
    assert set((bare_fs / build_opt.dst).iterdir()) == minimal_dst_top_level


def test_does_not_copy_symlinks(bare_fs, build_opt, minimal_dst_top_level):
    (bare_fs / build_opt.src / "link.lnk").symlink_to("/tmp")
    build(build_opt)
    assert set((bare_fs / build_opt.dst).iterdir()) == minimal_dst_top_level


def test_does_not_copy_destination_files(bare_fs, build_opt, minimal_dst_top_level):
    dst_path = bare_fs / build_opt.dst / "existing.html"
    make_fs(
        {
            dst_path: "<html></html>",
        }
    )
    build(build_opt)
    expected = minimal_dst_top_level | {dst_path}
    assert set((bare_fs / build_opt.dst).iterdir()) == expected


def test_does_not_copy_explicitly_skipped_files(
    bare_fs, build_opt, minimal_dst_top_level
):
    config_file = bare_fs / build_opt.config
    make_fs(
        {
            config_file: '[tool.mccole]\nskips = ["*.text", "extras/**", "uv.lock"]\n',
            bare_fs / build_opt.src / "alpha.text": "something",
            bare_fs / build_opt.src / "uv.lock": "version = 1",
            bare_fs / build_opt.src / "extras" / "test.md": "# test",
        }
    )
    build(build_opt)
    assert set((bare_fs / build_opt.dst).iterdir()) == minimal_dst_top_level


def test_footer_title_filled_in_correctly(bare_fs, build_opt, readme_dst_file):
    build(build_opt)
    doc = read_doc(readme_dst_file)
    footer = doc.select_one("footer")
    assert footer is not None
    links = {node["href"]: node.get_text() for node in footer.select("a[href]")}
    assert "./" in links
    assert links["./"] == "Tutorial"
    assert "./#acknowledgments" in links


def test_boilerplate_files_correctly_renamed(bare_fs, build_opt):
    fixtures = (
        ("CODE_OF_CONDUCT.md", "Code of Conduct", "conduct"),
        ("CONTRIBUTING.md", "Contributing", "contrib"),
        ("LICENSE.md", "License", "license"),
    )
    for filename, content, _ in fixtures:
        x = bare_fs / build_opt.src / filename
        x.write_text(f"# {content}\n")

    build(build_opt)

    for filename, content, output in fixtures:
        expected = bare_fs / build_opt.dst / output / "index.html"
        assert expected.is_file()
        assert content in expected.read_text()


def test_boilerplate_links_correctly_adjusted(bare_fs, build_opt):
    text = """\
# Title
<section id="text" markdown="1">
[conduct](./CODE_OF_CONDUCT.md)
[contributing](./CONTRIBUTING.md)
[license](./LICENSE.md)
[home page](./README.md)
</section>
"""
    make_fs({bare_fs / build_opt.src / "test.md": text})
    build(build_opt)

    expected = bare_fs / build_opt.dst / "test.html"
    assert expected.is_file()
    doc = read_doc(expected)
    section = doc.select("section[id='text']")[0]
    urls = {node["href"] for node in section.select("a")}
    assert urls == {"./conduct/", "./contrib/", "./license/", "./"}


def test_bibliography_links_correctly_adjusted(bare_fs, build_opt):
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": "# Title\n[](b:first)\n",
        }
    )
    build(build_opt)
    expected = bare_fs / build_opt.dst / "test.html"
    assert expected.is_file()
    assert 'href="./bibliography/#first"' in expected.read_text()


def test_glossary_links_correctly_adjusted(bare_fs, build_opt):
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": "# Title\n[term](g:key)\n",
        }
    )
    build(build_opt)
    expected = bare_fs / build_opt.dst / "test.html"
    assert expected.is_file()
    assert 'href="./glossary/#key"' in expected.read_text()


def test_glossary_keys_values_loaded(bare_fs, build_opt, glossary_src_file):
    glossary = _load_glossary([glossary_src_file])
    assert glossary == {"first": "first term", "second": "second term"}


def test_defined_terms_added_to_page(bare_fs, build_opt, glossary_dst_file):
    text = """\
# Title
<p id="terms"></p>
[one](g:first) [two](g:second)
"""
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)
    doc = read_doc(bare_fs / build_opt.dst / "test.html")
    terms = doc.select("p#terms")[0]
    refs = {node["href"] for node in terms.select("a[href]")}
    assert refs == {"./glossary/#first", "./glossary/#second"}


def test_defined_terms_paragraph_removed(bare_fs, build_opt):
    text = """\
# Title
<p id="terms"></p>
body
"""
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)
    expected = bare_fs / build_opt.dst / "test.html"
    assert expected.is_file()
    doc = read_doc(expected)
    assert not doc.select("p#terms")


def test_backtick_code_block_class_applied_to_enclosing_pre(bare_fs, build_opt):
    text = """\
# Title
```py
x = 1
```
"""
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)
    expected = bare_fs / build_opt.dst / "test.html"
    assert expected.is_file()
    doc = read_doc(expected)
    for tag in ["pre", "code"]:
        nodes = doc.find_all(tag)
        assert len(nodes) == 1
        assert nodes[0]["class"] == ["language-py"]


def test_non_markdown_files_copied(bare_fs, build_opt):
    make_fs(
        {
            bare_fs / build_opt.src / "in_root.txt": "root text",
            bare_fs / build_opt.src / "subdir" / "in_subdir.txt": "subdir text",
        }
    )
    build(build_opt)

    in_root = bare_fs / build_opt.dst / "in_root.txt"
    assert in_root.is_file()
    assert in_root.read_text() == "root text"

    in_subdir = bare_fs / build_opt.dst / "subdir" / "in_subdir.txt"
    assert in_subdir.is_file()
    assert in_subdir.read_text() == "subdir text"


def test_append_provided_links_file(bare_fs, build_opt):
    text = """\
# Title
[link][url]
"""
    links_path = bare_fs / build_opt.src / "links.txt"
    test_url = "http://some.url/"
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": text,
            links_path: f"[url]: {test_url}",
        }
    )
    build_opt.links = links_path
    build(build_opt)

    doc = read_doc(bare_fs / build_opt.dst / "test.html")
    links = doc.main.select("a")
    assert len(links) == 1
    assert links[0]["href"] == test_url


def test_labels_figures(bare_fs, build_opt):
    text = """
# Title
<figure id='f:first'><figcaption>caption</figcaption></figure>
"""
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)

    doc = read_doc(bare_fs / build_opt.dst / "test.html")
    captions = doc.select("figcaption")
    assert len(captions) == 1
    assert captions[0].string == "Figure 1: caption"


def test_cross_references_figures(bare_fs, build_opt):
    text = """\
# Title
<figure id='f:first'><figcaption>caption</figcaption></figure>
text [](#f:first)
"""
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)

    doc = read_doc(bare_fs / build_opt.dst / "test.html")
    links = [node for node in doc.select("a[href]") if node["href"].startswith("#f:")]
    assert len(links) == 1
    assert links[0].string == "Figure 1"


def test_create_table_correctly(bare_fs, build_opt):
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": TABLE_MD,
        }
    )
    build(build_opt)

    doc = read_doc(bare_fs / build_opt.dst / "test.html")
    nodes = {}
    for selector in ["table", "thead", "tbody", "caption"]:
        found = doc.select(selector)
        assert len(found) == 1
        nodes[selector] = found[0]

    assert nodes["table"]["id"] == "t:tbl"
    assert nodes["caption"].string == "Table Title"

    assert len(nodes["thead"].select("tr")) == 1
    assert len(nodes["thead"].select("th")) == 2
    assert len(nodes["thead"].select("td")) == 0
    assert set(n.string for n in nodes["thead"].select("th")) == {"left", "right"}

    assert len(nodes["tbody"].select("tr")) == 2
    assert len(nodes["tbody"].select("th")) == 0
    assert len(nodes["tbody"].select("td")) == 4
    assert set(n.string for n in nodes["tbody"].select("td")) == {"1", "2", "3", "4"}


def test_table_cross_reference(bare_fs, build_opt):
    text = f"""\
{TABLE_MD}
text [](#t:tbl)
"""
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)
    doc = read_doc(bare_fs / build_opt.dst / "test.html")
    links = [node for node in doc.select("a[href]") if node["href"].startswith("#t:")]
    assert len(links) == 1
    assert links[0].string == "Table 1"


def test_fail_badly_formatted_internal_readme_link(bare_fs, build_opt):
    readme = build_opt.src / "README.md"
    content = readme.read_text()
    content = content.replace("(./", "(../")
    readme.write_text(content)
    with pytest.raises(AssertionError):
        build(build_opt)


def test_warn_unknown_markdown_links(bare_fs, build_opt, capsys):
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": "# Title\n[text](link.md)\n",
        }
    )
    build(build_opt)
    captured = capsys.readouterr()
    assert "unknown Markdown link" in captured.err


def test_warn_missing_h1(bare_fs, build_opt, capsys):
    make_fs({bare_fs / build_opt.src / "test.md": "text\n"})
    build(build_opt)
    captured = capsys.readouterr()
    assert "lacks H1 heading" in captured.err


def test_warn_missing_title(bare_fs, build_opt, capsys):
    template_path = bare_fs / build_opt.templates / "page.html"
    template = template_path.read_text()
    template = template.replace("<title></title>", "")
    template_path.write_text(template)

    make_fs(
        {
            bare_fs / build_opt.src / "test.md": "text\n",
        }
    )
    build(build_opt)
    captured = capsys.readouterr()
    assert "does not have <title> element" in captured.err


def test_warn_badly_formatted_config_file(bare_fs, build_opt, capsys):
    config_file = bare_fs / "pyproject.toml"
    make_fs({config_file: '[tool.missing]\nskips = ["extras", "uv.lock"]\n'})
    build(build_opt)
    captured = capsys.readouterr()
    assert "does not have 'tool.mccole'" in captured.err


def test_warn_overlap_renames_and_skips(bare_fs, build_opt, capsys):
    config_file = bare_fs / "pyproject.toml"
    make_fs(
        {
            config_file: '[tool.mccole]\nskips = ["LICENSE.md"]\n',
            bare_fs / build_opt.src / "LICENSE.md": "# License",
        }
    )
    build(build_opt)
    captured = capsys.readouterr()
    assert "overlap between skips and renames" in captured.err


def test_warn_multiple_term_paragraphs_in_doc(bare_fs, build_opt, capsys):
    text = """\
# Terms
<p id="terms"></p>
<p id="terms"></p>
"""
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)
    captured = capsys.readouterr()
    assert "terms paragraph appears multiple times" in captured.err


def test_warn_no_glossary_file_found(bare_fs, build_opt, glossary_src_file, capsys):
    glossary_src_file.unlink()
    glossary_src_file.parent.rmdir()
    build(build_opt)
    captured = capsys.readouterr()
    assert "no glossary found" in captured.err


def test_warn_multiple_glossary_files_found(bare_fs, build_opt, capsys):
    make_fs(
        {
            bare_fs / build_opt.src / "subdir" / "glossary" / "index.md": "# Second",
        }
    )
    build(build_opt)
    captured = capsys.readouterr()
    assert "multiple glossary files" in captured.err


def test_warn_figure_missing_id(bare_fs, build_opt, capsys):
    text = """\
# Title
<figure><figcaption>caption</figcaption></figure>
"""
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)
    captured = capsys.readouterr()
    assert "has no ID" in captured.err


def test_warn_figure_id_wrong_prefix(bare_fs, build_opt, capsys):
    text = """\
# Title
<figure id='something'><figcaption>caption</figcaption></figure>
"""
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)
    captured = capsys.readouterr()
    assert "does not start with" in captured.err


def test_warn_figure_multiple_captions(bare_fs, build_opt, capsys):
    text = """
# Title
<figure id='f:fig'><figcaption>caption</figcaption><figcaption>another</figcaption></figure>
"""
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)
    captured = capsys.readouterr()
    assert "has missing/too many captions" in captured.err


def test_warn_figure_missing_caption(bare_fs, build_opt, capsys):
    text = """\
# Title
<figure id='f:fig'></figure>
"""
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)
    captured = capsys.readouterr()
    assert "has missing/too many captions" in captured.err


def test_warn_figure_reference_cannot_be_resolved(bare_fs, build_opt, capsys):
    text = """\
# Title
<figure id='f:fig'><figcaption>caption</figcaption></figure>
text [](#f:missing)
"""
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)
    captured = capsys.readouterr()
    assert "cannot resolve figure reference" in captured.err


def test_warn_table_id_wrong_prefix(bare_fs, build_opt, capsys):
    text = TABLE_MD.replace("t:tbl", "something")
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)
    captured = capsys.readouterr()
    assert "does not start with 't:'" in captured.err


def test_warn_table_missing_caption(bare_fs, build_opt, capsys):
    text = TABLE_MD.replace("data-table-caption", "data-something")
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)
    captured = capsys.readouterr()
    assert "does not have data-table-caption" in captured.err


def test_warn_table_no_table_element(bare_fs, build_opt, capsys):
    text = """\
# Title

<div markdown="1" data-table-id="t:tbl" data-table-caption="Table Title">
<p>Not a table</p>
</div>
"""
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)
    captured = capsys.readouterr()
    assert "does not contain table" in captured.err


def test_warn_table_reference_cannot_be_resolved(bare_fs, build_opt, capsys):
    text = f"""\
{TABLE_MD}
text [](#t:missing)
"""
    make_fs(
        {
            bare_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)
    captured = capsys.readouterr()
    assert "cannot resolve table reference" in captured.err



"""Generate .epub from output HTML."""

import argparse
from pathlib import Path
from datetime import datetime

from bs4 import BeautifulSoup
from ebooklib import epub


SECTIONS = ("div#lessons", "div#appendices")

NAV_CONTENT = """<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xmlns:epub="http://www.idpf.org/2007/ops">
<head>
    <title>Table of Contents</title>
</head>
<body>
    <nav epub:type="toc">
        <h1>Table of Contents</h1>
        <ol>
            {nav_items}
        </ol>
    </nav>
</body>
</html>"""

XHTML_TEMPLATE = """<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
    <title>{title}</title>
    {css_link}
</head>
<body>
<div>
{body}
</div>
</body>
</html>"""


def main(opt):
    """Generate .epub."""
    book_title, contents = _get_contents(opt)

    book = epub.EpubBook()
    book.set_identifier(opt.id)
    book.set_title(book_title)
    book.set_language("en")
    book.add_metadata("DC", "publisher", "Open Source")
    book.add_metadata("DC", "date", datetime.now().strftime("%Y-%m-%d"))
    for author in opt.authors:
        book.add_author(author)

    css_item = epub.EpubItem(
        uid="main_css",
        file_name="styles/main.css",
        media_type="text/css",
        content=opt.css.read_bytes(),
    )
    book.add_item(css_item)
    css_link = '<link rel="stylesheet" type="text/css" href="styles/main.css"/>'

    chapters = []
    spine = ["nav"]

    for i, (chapter_title, chapter_body) in enumerate(contents.items()):
        chapter = epub.EpubHtml(title=chapter_title, file_name=f"chapter_{i}.xhtml")
        chapter.content = XHTML_TEMPLATE.format(
            title=chapter_title, body=chapter_body.prettify(), css_link=css_link
        ).encode("utf-8")
        book.add_item(chapter)
        chapters.append(chapter)
        spine.append(chapter)

    book.toc = [
        (epub.Link(chap.file_name, chap.title, chap.id), []) for chap in chapters
    ]
    book.add_item(epub.EpubNcx())
    nav = epub.EpubNav()
    nav.content = NAV_CONTENT.format(
        nav_items="".join(
            f'<li><a href="{chap.file_name}">{chap.title}</a></li>' for chap in chapters
        )
    )
    book.add_item(nav)

    book.spine = spine
    epub.write_epub(opt.out, book)


def construct_parser(parser):
    """Parse command-line arguments."""
    parser.add_argument("--authors", nargs="+", help="author names")
    parser.add_argument("--css", type=Path, help="CSS file to include in EPUB")
    parser.add_argument("--id", type=str, help="book identifier")
    parser.add_argument("--out", type=Path, help="output file")
    parser.add_argument(
        "--src", type=Path, default=Path("docs"), help="input directory"
    )


def _get_contents(opt):
    """Get filepaths and titles from index.html."""
    index = opt.src / "index.html"
    doc = BeautifulSoup(index.read_text(), "html.parser")
    book_title = doc.select_one("h1").get_text().strip()
    contents = {}
    for selector in SECTIONS:
        div = doc.select_one(selector)
        assert div is not None, f"{index} does not contain {selector}"
        for entry in div.select("li>a[href]"):
            chapter_title = entry.get_text().strip()
            sub_path = opt.src / entry["href"] / "index.html"
            sub_doc = BeautifulSoup(sub_path.read_text(), "html.parser").select_one(
                "main"
            )
            assert sub_doc is not None, f"{sub_path} does not contain main section"
            assert sub_doc.get_text().strip(), f"{sub_path} main section is empty"
            contents[chapter_title] = sub_doc
    return book_title, contents



"""Refresh file inclusions in place."""

import argparse
from pathlib import Path
import re


COMMENT = {
    "js": "//",
    "py": "#",
    "sql": "--",
}

INC_PAT = re.compile(
    r'^```\s*\{.*?data-file="(.+?)".*?\}.+?^```\s*?$', re.DOTALL + re.MULTILINE
)

INCLUDE_BLOCK = """\
```{{data-file="{path}"}}
{content}\
```\
"""

INCLUDE_BLOCK_KEEP = """\
```{{data-file="{path}:{keep}"}}
{content}\
```\
"""


def main(opt):
    """Refresh inclusions in source file."""
    for filename in opt.files:
        src_path = Path(filename)
        content = src_path.read_text()
        content = INC_PAT.sub(lambda m: _do_inclusion(src_path, m), content)
        src_path.write_text(content)


def construct_parser(parser):
    """Parse command-line arguments."""
    parser.add_argument("--files", nargs="+", help="files to refresh")


def _do_inclusion(outer, match_obj):
    """Refresh content of a single inclusion."""
    spec = match_obj.group(1)
    inner, keep, content = _get_spec(outer, spec)
    if keep is None:
        return INCLUDE_BLOCK.format(path=inner, content=content)
    else:
        content = _extract_section(outer, inner, keep, content)
        return INCLUDE_BLOCK_KEEP.format(path=inner, content=content, keep=keep)


def _extract_section(outer, inner, keep, content):
    """Keep a section of a file."""
    suffix = Path(inner).suffix.lstrip(".")
    assert suffix in COMMENT, f"{outer}: unknown inclusion suffix in {inner}"
    before = f"{COMMENT[suffix]} [{keep}]"
    after = f"{COMMENT[suffix]} [/{keep}]"
    assert (before in content) and (after in content), (
        f"{outer} :: {inner}: missing start/end for {COMMENT[suffix]} and {keep}"
    )
    content = content.split(before)[1].split(after)[0]
    if content[0] == "\n":
        content = content[1:]
    return content


def _get_spec(outer, inner):
    """Load external included file."""
    if ":" in inner:
        inner, keep = inner.split(":")
    else:
        keep = None
    path = outer.parent / inner
    assert path.is_file(), (
        f"Bad inclusion in {outer}: {path} does not exist or is not file"
    )
    return inner, keep, path.read_text()



"""Test site checker."""

import argparse
import re

from bs4 import BeautifulSoup
import pytest

from mccole.build import main as build, construct_parser as build_parser
from mccole.lint import main as lint, construct_parser as lint_parser
from .conftest import make_fs

DIV_APPENDICES = re.compile(r'<div id="appendices">.+?</div>', re.DOTALL)
NAV_LESSONS = re.compile(
    '<span class="dropdown-content" id="nav-lessons">.+?</span>', re.DOTALL
)

MINIMAL_GLOSSARY_REFS = """
# Title

[first](g:first)
[second](g:second)
"""


@pytest.fixture
def lint_fs(build_opt, bare_fs):
    make_fs(
        {
            bare_fs / "bibliography" / "index.md": "# Bibliography",
        }
    )
    return bare_fs


def test_construct_parser_with_default_values():
    parser = argparse.ArgumentParser()
    lint_parser(parser)
    opt = parser.parse_args([])
    assert hasattr(opt, "dst")


def test_no_problems_to_report(build_opt, lint_opt, lint_fs, capsys):
    make_fs(
        {
            lint_fs / build_opt.src / "test.md": MINIMAL_GLOSSARY_REFS,
        }
    )
    build(build_opt)
    output_path = lint_fs / build_opt.dst / "test.html"
    lint(lint_opt)
    captured = capsys.readouterr()
    print("ERR", captured.err)
    assert not captured.err


def test_multiple_h1_in_file(build_opt, lint_opt, lint_fs, capsys):
    make_fs(
        {
            lint_fs / "test.md": "# First\n\n# Second\n",
        }
    )
    build(build_opt)
    lint(lint_opt)
    captured = capsys.readouterr()
    m = re.search(r"(\d) H1 elements found", captured.err)
    assert m and int(m.group(1)) == 2


@pytest.mark.parametrize("kind", ["bibliography", "glossary"])
def test_special_file_not_found(build_opt, lint_opt, lint_fs, capsys, kind):
    build(build_opt)
    (lint_fs / lint_opt.dst / kind / "index.html").unlink()
    lint(lint_opt)
    captured = capsys.readouterr()
    assert re.search(f"{kind}/index.html not found", captured.err)


@pytest.mark.parametrize("kind,key", [["bibliography", "b:key"], ["glossary", "g:key"]])
def test_special_key_undefined(build_opt, lint_opt, lint_fs, capsys, kind, key):
    text = f"""\
# A
[]({key})
"""
    make_fs(
        {
            lint_fs / "a.md": text,
        }
    )

    build(build_opt)
    lint(lint_opt)

    captured = capsys.readouterr()
    assert re.search(f"unknown {kind} keys in .+: key", captured.err)


@pytest.mark.parametrize("kind", ["bibliography", "glossary"])
def test_special_key_unused(build_opt, lint_opt, lint_fs, capsys, kind):
    build(build_opt)

    text = """\
# Title
<main>
<span id="key">text</span>
</main>
"""
    make_fs(
        {
            lint_fs / lint_opt.dst / kind / "index.html": text,
        }
    )

    lint(lint_opt)

    captured = capsys.readouterr()
    assert re.search(f"unused {kind} keys: key", captured.err)


def test_multiple_main_in_special_file(build_opt, lint_opt, lint_fs, capsys):
    build(build_opt)
    page = build_opt.dst / "bibliography" / "index.html"
    content = page.read_text()
    content = content.replace("<main>", "<main></main><main>")
    page.write_text(content)

    lint(lint_opt)
    captured = capsys.readouterr()
    assert "missing or multiple" in captured.err


def test_glossary_term_redefined(build_opt, lint_opt, lint_fs, capsys):
    a = """\
# A
[term 1](g:term_1)
[term 1 again](g:term_1)
[term 2](g:term_2)
"""
    b = """\
# B
[term 2](g:term_2)
"""
    g = """\
# Glossary
<span id="term_1">term 1</span>
<span id="term_2">term 2</span>
"""
    make_fs(
        {
            lint_fs / "a.md": a,
            lint_fs / "b.md": b,
            lint_fs / "glossary" / "index.md": g,
        }
    )

    build(build_opt)
    lint(lint_opt)

    captured = capsys.readouterr()
    assert re.search(r"glossary entry 'term_1' defined in .+, .+", captured.err)
    assert re.search(r"glossary entry 'term_2' defined in .+, .+", captured.err)


def test_glossary_key_unused(build_opt, lint_opt, lint_fs, capsys):
    build(build_opt)

    g = """\
# Glossary
<main>
<span id="key">text</span>
</main>
"""
    make_fs(
        {
            lint_fs / lint_opt.dst / "glossary" / "index.html": g,
        }
    )

    lint(lint_opt)

    captured = capsys.readouterr()
    assert re.search("unused glossary keys: key", captured.err)


@pytest.mark.parametrize(
    "headings,msg",
    [
        ["", "has missing/multiple heading(s)"],
        ["## first\n## second", "has missing/multiple heading(s)"],
        ["# level 1", "has h1 instead of h3"],
        ["## level 2", "has h2 instead of h3"],
        ["#### level 4", "has h4 instead of h3"],
    ],
)
def test_exercise_section_has_bad_headings(
    build_opt, lint_opt, lint_fs, capsys, headings, msg
):
    text = f"""\
# Title
<section class="exercise" markdown="1">
{headings}
</section>
"""
    make_fs(
        {
            lint_fs / build_opt.src / "test.md": text,
        }
    )

    build(build_opt)
    lint(lint_opt)

    captured = capsys.readouterr()
    assert msg in captured.err


def test_html_validation_with_valid_html(build_opt, lint_opt, lint_fs, capsys):
    make_fs(
        {
            lint_fs / build_opt.src / "test.md": MINIMAL_GLOSSARY_REFS,
        }
    )
    build(build_opt)
    lint_opt.html = True
    lint(lint_opt)
    captured = capsys.readouterr()
    assert not captured.err


def test_html_validation_with_invalid_html(build_opt, lint_opt, lint_fs, capsys):
    build(build_opt)
    page = build_opt.dst / "index.html"
    content = page.read_text()
    page.write_text(content.replace('<html lang="en">', ""))

    lint_opt.html = True
    lint(lint_opt)
    captured = capsys.readouterr()
    assert captured.out


@pytest.mark.parametrize(
    "pattern,replacement",
    [
        [NAV_LESSONS, ""],
        [DIV_APPENDICES, ""],
    ],
)
def test_compare_template_readme_missing_nav(
    build_opt, lint_opt, lint_fs, capsys, pattern, replacement
):
    build(build_opt)
    page = build_opt.dst / "index.html"
    content = page.read_text()
    content = pattern.sub(replacement, content)
    page.write_text(content)

    lint(lint_opt)
    captured = capsys.readouterr()
    assert "missing or multiple" in captured.err


def test_compare_template_readme_mismatch_titles(build_opt, lint_opt, lint_fs, capsys):
    build(build_opt)
    page = build_opt.dst / "index.html"
    content = page.read_text()
    content = content.replace(
        '<span id="nav-lessons" class="dropdown-content"></span>',
        '<span id="nav-lessons" class="dropdown-content"><a href="./01_intro/">Introduction</a></span>',
    )
    content = content.replace(
        '<div id="lessons"></div>',
        '<div id="lessons"><li><a href="./01_intro/">Not Introduction</a></li></div>',
    )
    page.write_text(content)

    lint(lint_opt)
    captured = capsys.readouterr()
    assert "mis-match in README and nav paths" in captured.err


def test_figure_missing_id(build_opt, lint_opt, lint_fs, capsys):
    text = """\
# Title
<figure>
<figcaption>Figure 1: A caption</figcaption>
</figure>
"""
    make_fs(
        {
            lint_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)
    lint(lint_opt)
    captured = capsys.readouterr()
    assert "figure missing 'id'" in captured.err


def test_figure_missing_caption(build_opt, lint_opt, lint_fs, capsys):
    text = """\
# Title
<figure id="f:test">
</figure>
"""
    make_fs(
        {
            lint_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)
    lint(lint_opt)
    captured = capsys.readouterr()
    assert "figure missing/extra caption" in captured.err


def test_figure_has_multiple_captions(build_opt, lint_opt, lint_fs, capsys):
    text = """\
# Title
<figure id="f:test">
<figcaption>Figure 1: First caption</figcaption>
<figcaption>Figure 1: Second caption</figcaption>
</figure>
"""
    make_fs(
        {
            lint_fs / build_opt.src / "test.md": text,
        }
    )
    build(build_opt)
    lint(lint_opt)
    captured = capsys.readouterr()
    assert "figure missing/extra caption" in captured.err


def test_figure_badly_formatted_caption(build_opt, lint_opt, lint_fs, capsys):
    build(build_opt)
    html_content = """\
<!DOCTYPE html>
<html>
<head><title>Test</title></head>
<body>
<main>
<figure id="f:test">
<figcaption>Bad caption format</figcaption>
</figure>
</main>
</body>
</html>
"""
    make_fs(
        {
            lint_fs / lint_opt.dst / "test.html": html_content,
        }
    )
    lint(lint_opt)
    captured = capsys.readouterr()
    assert "badly-formatted figure caption 'Bad caption format'" in captured.err


def test_table_missing_id(build_opt, lint_opt, lint_fs, capsys):
    build(build_opt)
    html_content = """\
<!DOCTYPE html>
<html>
<head><title>Test</title></head>
<body>
<main>
<div data-table-caption="Table Title">
<table>
<tr><td>data</td></tr>
</table>
</div>
</main>
</body>
</html>
"""
    make_fs(
        {
            lint_fs / lint_opt.dst / "test.html": html_content,
        }
    )
    lint(lint_opt)
    captured = capsys.readouterr()
    assert "table div missing 'data-table-id'" in captured.err


def test_table_wrong_id_prefix(build_opt, lint_opt, lint_fs, capsys):
    build(build_opt)
    html_content = """\
<!DOCTYPE html>
<html>
<head><title>Test</title></head>
<body>
<main>
<div data-table-id="wrong:test" data-table-caption="Table Title">
<table>
<tr><td>data</td></tr>
</table>
</div>
</main>
</body>
</html>
"""
    make_fs(
        {
            lint_fs / lint_opt.dst / "test.html": html_content,
        }
    )
    lint(lint_opt)
    captured = capsys.readouterr()
    assert "does not start with 't:'" in captured.err


def test_table_missing_caption(build_opt, lint_opt, lint_fs, capsys):
    build(build_opt)
    html_content = """\
<!DOCTYPE html>
<html>
<head><title>Test</title></head>
<body>
<main>
<div data-table-id="t:test">
<table>
<tr><td>data</td></tr>
</table>
</div>
</main>
</body>
</html>
"""
    make_fs(
        {
            lint_fs / lint_opt.dst / "test.html": html_content,
        }
    )
    lint(lint_opt)
    captured = capsys.readouterr()
    assert "table div missing 'data-table-caption'" in captured.err


def test_table_no_table_element(build_opt, lint_opt, lint_fs, capsys):
    build(build_opt)
    html_content = """\
<!DOCTYPE html>
<html>
<head><title>Test</title></head>
<body>
<main>
<div data-table-id="t:test" data-table-caption="Table Title">
<p>Not a table</p>
</div>
</main>
</body>
</html>
"""
    make_fs(
        {
            lint_fs / lint_opt.dst / "test.html": html_content,
        }
    )
    lint(lint_opt)
    captured = capsys.readouterr()
    assert "table div should contain exactly one table, found 0" in captured.err


def test_table_multiple_table_elements(build_opt, lint_opt, lint_fs, capsys):
    build(build_opt)
    html_content = """\
<!DOCTYPE html>
<html>
<head><title>Test</title></head>
<body>
<main>
<div data-table-id="t:test" data-table-caption="Table Title">
<table><tr><td>first</td></tr></table>
<table><tr><td>second</td></tr></table>
</div>
</main>
</body>
</html>
"""
    make_fs(
        {
            lint_fs / lint_opt.dst / "test.html": html_content,
        }
    )
    lint(lint_opt)
    captured = capsys.readouterr()
    assert "table div should contain exactly one table, found 2" in captured.err


def test_link_in_page_body_resolves(build_opt, lint_opt, lint_fs, capsys):
    markdown_content = (
        MINIMAL_GLOSSARY_REFS
        + """\

[text][url]

[url]: http://something
"""
    )
    make_fs(
        {
            lint_fs / lint_opt.src / "test.md": markdown_content,
        }
    )
    build(build_opt)
    lint(lint_opt)
    captured = capsys.readouterr()
    assert not captured.err


def test_unresolved_link_in_page_body(build_opt, lint_opt, lint_fs, capsys):
    markdown_content = (
        MINIMAL_GLOSSARY_REFS
        + """\

[text][url]
"""
    )
    make_fs(
        {
            lint_fs / lint_opt.src / "test.md": markdown_content,
        }
    )
    build(build_opt)
    lint(lint_opt)
    captured = capsys.readouterr()
    assert "undefined link reference(s)" in captured.err


def test_link_in_external_links(build_opt, lint_opt, lint_fs, capsys):
    markdown_content = (
        MINIMAL_GLOSSARY_REFS
        + """\

[text][url]
"""
    )
    links_content = """\
[url]: http://something
"""
    links_path = lint_fs / lint_opt.src / "links.txt"
    make_fs(
        {
            lint_fs / lint_opt.src / "test.md": markdown_content,
            links_path: links_content,
        }
    )
    build(build_opt)
    lint_opt.links = links_path
    lint(lint_opt)
    captured = capsys.readouterr()
    assert not captured.err


def test_unused_internal_link_definition(build_opt, lint_opt, lint_fs, capsys):
    markdown_content = (
        MINIMAL_GLOSSARY_REFS
        + """\

[url]: http://something
"""
    )
    make_fs(
        {
            lint_fs / lint_opt.src / "test.md": markdown_content,
        }
    )
    build(build_opt)
    lint(lint_opt)
    captured = capsys.readouterr()
    assert "unused link definition(s)" in captured.err


def test_unused_global_link_definition(build_opt, lint_opt, lint_fs, capsys):
    links_content = """\
[url]: http://something
"""
    links_path = lint_fs / lint_opt.src / "links.txt"
    make_fs(
        {
            lint_fs / lint_opt.src / "test.md": MINIMAL_GLOSSARY_REFS,
            links_path: links_content,
        }
    )
    build(build_opt)
    lint_opt.links = links_path
    lint(lint_opt)
    captured = capsys.readouterr()
    assert "unused global link definition(s)" in captured.err



"""Interface for command-line script."""

import argparse
import importlib.metadata
import sys

from .build import main as build, construct_parser as build_parser
from .epub import main as epub, construct_parser as epub_parser
from .init import main as init, construct_parser as init_parser
from .lint import main as lint, construct_parser as lint_parser
from .refresh import main as refresh, construct_parser as refresh_parser


COMMANDS = {
    "build": (build, build_parser, "build site"),
    "epub": (epub, epub_parser, "create epub"),
    "init": (init, init_parser, "(re-)initialize site"),
    "lint": (lint, lint_parser, "check site"),
    "refresh": (refresh, refresh_parser, "refresh inclusions"),
}


def main():
    """Main driver."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--version", action="store_true", help="show version")

    subparsers = parser.add_subparsers(dest="cmd")
    for cmd, (_, p, text) in COMMANDS.items():
        p(subparsers.add_parser(cmd, help=text))

    opt = parser.parse_args()
    if opt.version:
        print(importlib.metadata.version("mccole"))
    if opt.cmd in COMMANDS:
        COMMANDS[opt.cmd][0](opt)
    else:
        print(f"unknown command {opt.cmd}", file=sys.stderr)
        sys.exit(1)
